/////////////////////////////////////////////////////////////////////
//
//            X   X           X
//           XXX XX         XX
//          XXXXXXXX      XXX
//         XX X XXXXXXXXXXX
//        XXXXX XXXXXXXXX
//       XXXXX XXXXXXXXXX
//            XXX XXX XXX
//           XXX XX   XX
//           X   X     X
//
//    Copyright (C) 2003-2016  Ron Jakl
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////

#ifndef GRAPHWIDGETHEADERFILE
#define GRAPHWIDGETHEADERFILE

#include <QVariant>
#include <QWidget>
#include <QRect>
#include <QPainter>
#include <QPixmap>

#include <vector>

#include "vector2.h"

class QWidget;
class QPaintEvent;
class QResizeEvent;
class QMouseEvent;
class QPainter;

class TGraphWidget : public QWidget
{
	Q_OBJECT

public:

enum TGraphType
{
	DEFAULT_GRAPH=0,
	LINE_GRAPH,
	BAR_GRAPH,
	POINT_GRAPH
};

enum TLegendMode
{
	NO_LEGEND=0,
	DEFAULT_LEGEND_PRECISION,
	INTEGER_LEGEND_PRECISION
};


// CREATORS
	TGraphWidget(const QWidget *Parent=0,Qt::WindowFlags Flags=0);
	virtual ~TGraphWidget(void);

// ACCESSORS
	int NumberDataSets(void) const {return static_cast<int>(d_DataSets.size());}
	int DataSetSize(const int Set) const;
	TVector2 GetPoint(const int Set,const int Offset) const;

// MANIPULATORS	
	void VerticalLegend(const QString &s) {d_VLegendTitle = s;}
	void HorizontalLegend(const QString &s) {d_HLegendTitle = s;}
	void Title(const QString &s) {d_Title = s;}
	void SetHLegendMode(const TLegendMode State) {d_HLegendMode = State;}
	void SetVLegendMode(const TLegendMode State) {d_VLegendMode = State;}
	void SetGraphType(const TGraphType Type) {d_GraphType = Type;}

	bool DeleteDataSet(const int Set);
	void DeleteDataSets(void);

	int CreateDataSet(const QColor Color,const TGraphWidget::TGraphType = TGraphWidget::DEFAULT_GRAPH);
	bool AddPoint(const int Set,const TVector2&);
	bool ModifyPoint(const int Set,const int Offset,const TVector2&);
	bool DeletePoint(const int Set,const int Offset);
	
	void UpdateGraph(void);
	
	bool PaintToDevice(QPainter *Device,const QRect &DeviceRect,const bool draw_border = true);
	bool PaintToDevice_GraphOnly(QPainter *Device,const QRect &DeviceRect,const bool draw_border = true);
	
	void SetScaleMinimumY(const double yerr);
	void LockAspectRatio(const bool state);
	

// GenerateRGBColor always creates color.  Returns false and black when set is above 96
	static bool GenerateRGBColor(const int Set,int *R,int *G,int *B);
	

public slots:
	void ZoomOut(void);
	void ScaleToFit(void);

protected:
	
	virtual void paintEvent(QPaintEvent*);
	virtual void resizeEvent(QResizeEvent*);
	virtual void mousePressEvent(QMouseEvent*);
	virtual void mouseMoveEvent(QMouseEvent*);
	virtual void mouseReleaseEvent(QMouseEvent*);
	
private:

// NOT IMPLEMENTED
	TGraphWidget(const TGraphWidget&);
	TGraphWidget& operator=(const TGraphWidget&);	

struct TViewportMat
{
	double								tx[2];
	double								ty[2];
	double								to[2];

	TViewportMat& operator*=(const TViewportMat &rhs)
	{
		double							x,y,o[2];

		x = tx[0] * rhs.tx[0];
		y = ty[1] * rhs.ty[1];
		o[0] = to[0] * rhs.tx[0] + rhs.to[0];
		o[1] = to[1] * rhs.ty[1] + rhs.to[1];

		tx[0] = x;
		ty[1] = y;
		to[0] = o[0];
		to[1] = o[1];

		return *this;
	}

	void TranslateToWin(const double Xd, const double Yd, int *X, int *Y) const
	{
		*X = static_cast<int>(Xd * tx[0] + to[0]);
		*Y = static_cast<int>(Yd * ty[1] + to[1]);
	}

	void TranslateFromWin(const int X, const int Y, double *Xd, double *Yd) const
	{
		*Xd = (static_cast<double>(X) - to[0]) / tx[0];
		*Yd = (static_cast<double>(Y) - to[1]) / ty[1];
	}
	
};

struct TDataSet
{
	std::vector<TVector2>				Points;
	QColor								Color;
	TGraphType							DatasetGraphType;
};

	TViewportMat						d_ViewportMat;
	std::vector<TDataSet>				d_DataSets;

	QRect								d_GraphRect;
	QRect								d_VLegendRect;
	QRect								d_VLegendTitleRect;
	QRect								d_HLegendRect;
	QRect								d_HLegendTitleRect;
	QRect								d_TitleRect;
	QRect								d_Viewport;

	QString								d_VLegendTitle;
	QString								d_HLegendTitle;
	QString								d_Title;
	QString								d_VFormat;
	QString								d_HFormat;
	
	QSize								d_VMaxLabelSize;
	QSize								d_HMaxLabelSize;
	int									d_LineHeight;
	TGraphType							d_GraphType;

	double								d_RangeX;
	double								d_RangeY;
	QPixmap								d_Pixmap;

	QString								d_CoordDisplay;

	QRect								d_MouseDownBox;
	bool								d_ShowMouseDownBox;
	bool								d_LeftMouseDown;
	bool								d_LeftMouseDownMoving;
	bool								d_RightMouseDown;
	bool								d_RightMouseDownMoving;
	
	bool								d_draw_border;
	bool								d_ScaleToFitOnPaint;
	bool								d_ZoomOutOnPaint;
	TLegendMode							d_HLegendMode;
	TLegendMode							d_VLegendMode;
	
	double								d_y_minmax;
	bool								d_lock_aspect;
	
	
	void DrawPixmap(void);
	void DrawPixmap(const QRect &Rect,const bool Resize=false);	// Resize pixmap.  Faster Pan
	void DrawPixmap(QPainter *Device);
	void CalculateViewport(void);
	void CalculateViewport(const QRect &Rect);
	void DrawLegend(QPainter*);
	void DrawGraph(QPainter*);
	void CalculateLegendFormat(QPainter*);
	void UpdateLeftDownBox(void);
	
};

#endif
