/////////////////////////////////////////////////////////////////////
//
//            X   X           X
//           XXX XX         XX
//          XXXXXXXX      XXX
//         XX X XXXXXXXXXXX
//        XXXXX XXXXXXXXX
//       XXXXX XXXXXXXXXX
//            XXX XXX XXX
//           XXX XX   XX
//           X   X     X
//
//    Copyright (C) 2003-2016  Ron Jakl
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////

#include <QApplication>
#include <math.h>
#include <assert.h>

#include "encoderchannel.h"

TEncoderChannel::TEncoderChannel(void)
:	d_point_count(100),
	d_gain(1.0),
	d_offset(0.0),
	d_phase(0.0)
{
}

TEncoderChannel::~TEncoderChannel(void)
{
}

double TEncoderChannel::Signal_Data(
	const double						&angle) const
{
	double								data;
	
	if(d_channel_mode == TEncoderChannel::MODE_SINE)
	{
		data = sin(angle + d_phase);
		data *= d_gain;
		data += d_offset;
		
		// clipping
		if(data > 1.0)
		{
			data = 1.0;
		}
		else if(data < -1.0)
		{
			data = -1.0;
		}
	}
	else if(d_channel_mode == TEncoderChannel::MODE_COSINE)
	{
		data = cos(angle + d_phase);
		data *= d_gain;
		data += d_offset;
		
		// clipping
		if(data > 1.0)
		{
			data = 1.0;
		}
		else if(data < -1.0)
		{
			data = -1.0;
		}
	}
	else
	{
		assert(false);	// should never be here
	}

	return data;
}

void TEncoderChannel::Set_Gain(
	const double						&gain)
{
	d_gain = gain;
}

void TEncoderChannel::Set_Offset(
	const double						&offset)
{
	d_offset = offset;
}

void TEncoderChannel::Set_Phase(
	const double						&phase)
{
	d_phase = phase;
}

double TEncoderChannel::Peak_Positive_Level(void) const
{
	double								peak(1.0);
	
	peak *= d_gain;
	peak += d_offset;
	
	if(peak > 1.0)
	{
		peak = 1.0;
	}
	
	return peak;
}

double TEncoderChannel::Peak_Negative_Level(void) const
{
	double								peak(-1.0);
	
	peak *= d_gain;
	peak += d_offset;
	
	if(peak < -1.0)
	{
		peak = -1.0;
	}
	
	return peak;
}

void TEncoderChannel::Update_Signals(void)
{
	const double						range(6.28318530717959);
	const double						data_cycles(4);
	double								incr;
	double								angle;
	int									cntr;
	TVector2							pnt;
	int									data_count;
	
	assert(d_point_count > 1);
	
	incr = range / static_cast<double>(d_point_count - 1);
	
	data_count = d_point_count * data_cycles;
	
	d_signal_data.clear();
	d_signal_data.resize(data_count);
	
	if(d_channel_mode == TEncoderChannel::MODE_SINE)
	{
		for(cntr = 0, angle = 0.0;cntr < data_count;++cntr, angle += incr)
		{
			pnt.x = angle;
			
			pnt.y = sin(pnt.x + d_phase);
			pnt.y *= d_gain;
			pnt.y += d_offset;
			
			// clipping
			if(pnt.y > 1.0)
			{
				pnt.y = 1.0;
			}
			else if(pnt.y < -1.0)
			{
				pnt.y = -1.0;
			}
			
			d_signal_data[cntr] = pnt;
		}
	}
	else if(d_channel_mode == TEncoderChannel::MODE_COSINE)
	{
		for(cntr = 0, angle = 0.0;cntr < data_count;++cntr, angle += incr)
		{
			pnt.x = angle;
			
			pnt.y = cos(pnt.x + d_phase);
			pnt.y *= d_gain;
			pnt.y += d_offset;
			
			// clipping
			if(pnt.y > 1.0)
			{
				pnt.y = 1.0;
			}
			else if(pnt.y < -1.0)
			{
				pnt.y = -1.0;
			}
	
			d_signal_data[cntr] = pnt;
		}
	}
	else
	{
		assert(false);	// should never be here
	}
}
