/////////////////////////////////////////////////////////////////////
//
//            X   X           X
//           XXX XX         XX
//          XXXXXXXX      XXX
//         XX X XXXXXXXXXXX
//        XXXXX XXXXXXXXX
//       XXXXX XXXXXXXXXX
//            XXX XXX XXX
//           XXX XX   XX
//           X   X     X
//
//    Copyright (C) 2003-2016  Ron Jakl
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////

#include <QApplication>
#include <QString>

#include "encoderinterpolator.h"

TEncoderInterpolator::TEncoderInterpolator(void)
:	d_data_size(5),
	d_interpolator_count(10)
{
}

TEncoderInterpolator::~TEncoderInterpolator(void)
{
}

std::vector<TVector2> TEncoderInterpolator::Interpolate(
	const double						&distance,
	const TEncoderChannel				&cha,
	const TEncoderChannel				&chb,
	bool								*error) const
{
	const double						zero_epsilon(0.000000001);
	const double						pi2(6.28318530717959);
	std::vector<TVector2>				data;
	TVector2							pnt;
	int									cntr;
	int									sample_size;
	int									ival;
	double								distance_incr;
	double								distance_interpolator_incr;
	double								distance_sample_position;
	double								distance_interpolator;
	double								angle;
	double								angle_incr;
	double								angle_sample_position;
	double								angle_interpolator_incr;
	double								angle_interpolator;
	double								value_sin;
	double								value_cos;
	
	assert(d_data_size > 0);
	assert(d_interpolator_count > 0);
	assert(distance > zero_epsilon);
	assert(error);
	
	sample_size = d_data_size;
	sample_size *= d_interpolator_count;
	
	*error = false;
	
	angle_incr = pi2 / static_cast<double>(sample_size);
	angle_interpolator_incr = pi2 / static_cast<double>(d_interpolator_count);
	
	distance_incr = distance / static_cast<double>(sample_size);
	distance_interpolator_incr = distance / static_cast<double>(d_interpolator_count);
	
	// starting at half step below zero
	angle_sample_position = -1 * angle_incr / 2.0;

	// forward direction.  angular rotation is ccw
	for(cntr = 0;cntr < sample_size;++cntr)
	{
		value_sin = cha.Signal_Data(angle_sample_position);
		value_cos = chb.Signal_Data(angle_sample_position);
		
		angle = this->Convert_Sin_Cos_To_Angle(value_sin,value_cos);
		
		if(cntr == 0)
		{
			// initialize starting position
			angle_interpolator = angle;
			distance_interpolator = 0.0;
			distance_sample_position = distance_interpolator - distance_incr * 2;
			
		}
		else
		{
			ival = this->Compare_Angles(angle,angle_interpolator,angle_interpolator_incr);
			
			if(ival < 0)
			{
				angle_interpolator -= angle_interpolator_incr;
				distance_interpolator -= distance_interpolator_incr;
			}
			else if(ival > 0)
			{
				angle_interpolator += angle_interpolator_incr;
				distance_interpolator += distance_interpolator_incr;
			}
		}
		
		pnt.x = distance_sample_position;
		pnt.y = distance_sample_position - distance_interpolator;
		
		data.push_back(pnt);
		
		angle_sample_position += angle_incr;
		distance_sample_position += distance_incr;
		
	}
	
	// check for lost count
	// final position must be within one distance_incr from the nominal length
	if(fabs(distance_interpolator - distance) > (distance_interpolator_incr + zero_epsilon))
	{
		*error = true;
	}

	// reverse direction.  angular rotation is cw
	for(cntr = 0;cntr < sample_size;++cntr)
	{
		value_sin = cha.Signal_Data(angle_sample_position);
		value_cos = -1 * chb.Signal_Data(angle_sample_position);
		
		angle = this->Convert_Sin_Cos_To_Angle(value_sin,value_cos);
		
		if(cntr == 0)
		{
			// initialize starting position
			angle_interpolator = angle;
			distance_sample_position = distance_interpolator + distance_incr * 2;
		}
		else
		{
			ival = this->Compare_Angles(angle,angle_interpolator,angle_interpolator_incr);
			
			if(ival < 0)
			{
				angle_interpolator -= angle_interpolator_incr;
				distance_interpolator += distance_interpolator_incr;
			}
			else if(ival > 0)
			{
				angle_interpolator += angle_interpolator_incr;
				distance_interpolator -= distance_interpolator_incr;
			}
		}
		
		pnt.x = distance_sample_position;
		pnt.y = distance_sample_position - distance_interpolator;
		
		data.push_back(pnt);
		
		angle_sample_position -= angle_incr;
		distance_sample_position -= distance_incr;
	}
	
	angle_sample_position += angle_incr;
	
	if((*error) == true ||
	   fabs(distance_interpolator) > (distance_interpolator_incr + zero_epsilon))
	{
		*error = true;
	}

	
	return data;
}

double TEncoderInterpolator::Convert_Sin_Cos_To_Angle(
	const double						&in_sin,
	const double						&in_cos) const
{
	const double						pi(3.141592653589793);
	const double						zero_epsilon(0.0000000001);
	double								return_value;
	double								value_sin(in_sin);
	double								value_cos(in_cos);
	double								value_tan;
	
	
	// clip and clean
	if(value_sin > 1.0)
	{
		value_sin = 1.0;
	}
	else if(value_sin < -1.0)
	{
		value_sin = -1.0;
	}
	
	if(value_cos > 1.0)
	{
		value_cos = 1.0;
	}
	else if(value_cos < -1.0)
	{
		value_cos = -1.0;
	}
	
	// check cosine value.  cannot be zero
	// if cosine value is zero the value is approximated
	// using a very large number with the proper sign
	if(fabs(value_cos) < zero_epsilon)
	{
		if(value_cos < 0)
		{
			value_cos = -1 * zero_epsilon;
		}
		else
		{
			value_cos = zero_epsilon;
		}
	}
	
	// convert to tan
	value_tan = value_sin / value_cos;
	return_value = atan(value_tan);
	
	// check for angle over pi/2
	if(value_cos < 0.0)
	{
		return_value += pi;
	}
	else if(value_sin < 0.0)
	{
		return_value += pi + pi;
	}

	return return_value;
}

int TEncoderInterpolator::Compare_Angles(
	const double						&angle,
	const double						&reference_angle,
	const double						&threshold) const
{
	const double						pi2(6.28318530717959);
	const double						zero_epsilon(0.000000001);
	int									return_value(0);
	double								compare_angle_negative;
	double								compare_reference_angle_negative;
	double								compare_angle(angle);
	double								compare_reference_angle(reference_angle);
	double								compare_threshold(fabs(threshold));
	
	// shift all angles between 0 and 2pi
	
	while(compare_angle > pi2)
	{
		compare_angle -= pi2;
	};
	
	while(compare_angle < 0.0)
	{
		compare_angle += pi2;
	}
	
	while(compare_reference_angle > pi2)
	{
		compare_reference_angle -= pi2;
	};
	
	while(compare_reference_angle < 0.0)
	{
		compare_reference_angle += pi2;
	}

	compare_angle_negative = compare_angle - pi2;
	compare_reference_angle_negative = compare_reference_angle - pi2;
	
	// find closes compare angle to reference angle.
	// 2pi and zero are the same
	if(fabs(compare_angle - compare_reference_angle_negative) < fabs(compare_angle - compare_reference_angle))
	{
		compare_reference_angle = compare_reference_angle_negative;
	}
	
	if(fabs(compare_reference_angle - compare_angle_negative) < fabs(compare_reference_angle - compare_angle))
	{
		compare_angle = compare_angle_negative;
	}
	
	if(compare_angle > (compare_reference_angle + compare_threshold - zero_epsilon))
	{
		return_value = 1;
	}
	else if(compare_angle < (compare_reference_angle - compare_threshold + zero_epsilon))
	{
		return_value = -1;
	}

	return return_value;
}

